/*
 * Argus Software
 * Copyright (c) 2000-2007 QoSient, LLC
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/* 
 * $Id: //depot/argus/argus-3.0/clients/radium/radium.c#29 $
 * $DateTime: 2006/03/31 15:01:50 $
 * $Change: 797 $
 */

/*
 * radium.c  - this is the argus record distribtion node.
 *    Acting just like a ra* program, supporting all the options
 *    and functions of ra(), and providing access to data, like
 *    argus, supporting remote filtering, and MAR record generation.
 *    This is an important workhorse for the argus architecture.
 */


#include <unistd.h>
#include <stdlib.h>

#include <rabins.h>
#include <radium.h>

int RaRealTime = 0;
float RaUpdateRate = 1.0;
                                                                                                                           
struct timeval ArgusLastRealTime = {0, 0};
struct timeval ArgusLastTime     = {0, 0};
struct timeval ArgusThisTime     = {0, 0};
                                                                                                                           
struct timeval dLastTime = {0, 0};
struct timeval dRealTime = {0, 0};
struct timeval dThisTime = {0, 0};
struct timeval dTime     = {0, 0};
                                                                                                                           
long long thisUsec = 0;
                                                                                                                           
void RadiumSendFile (struct ArgusOutputStruct *, struct ArgusClientData *, char *, int);

void
ArgusClientInit (struct ArgusParserStruct *parser)
{
   struct ArgusModeStruct *mode;
   struct timeval *tvp;
   int thread = 0;
   int pid, dflag;

#if defined(ARGUS_THREADS)
   thread++;
#endif /* ARGUS_THREADS */
 
   if (thread == 0)
      ArgusLog (LOG_ERR, "not compiled with pthread support.  exiting");

   parser->RaWriteOut = 1;
   parser->ArgusReverse = 1;

   if (!(parser->RaInitialized)) {
      (void) signal (SIGHUP,  (void (*)(int)) ArgusShutDown);
      (void) signal (SIGTERM, (void (*)(int)) ArgusShutDown);
      (void) signal (SIGQUIT, (void (*)(int)) ArgusShutDown);
      (void) signal (SIGINT,  (void (*)(int)) ArgusShutDown);
      (void) signal (SIGPIPE, SIG_IGN);

      if ((mode = parser->ArgusModeList) != NULL) {
         while (mode) {
            if (!(strncasecmp (mode->mode, "rtime", 5)) ||
               (!(strncasecmp (mode->mode, "realtime", 8)))) {
               char *ptr = NULL;
               RaRealTime++;
               if ((ptr = strchr(mode->mode, ':')) != NULL) {
                  double value = 0.0;
                  char *endptr = NULL;
                  ptr++;
                  value = strtod(ptr, &endptr);
                  if (ptr != endptr) {
                     RaUpdateRate = value;
                  }
               }
            }

            mode = mode->nxt;
         }
      }

      dflag = parser->dflag;

      if (parser->ArgusFlowModelFile != NULL) {
         RadiumParseResourceFile (parser, parser->ArgusFlowModelFile);
      } else {
         if (!(parser->Xflag))
            RadiumParseResourceFile (parser, "/etc/radium.conf");
      }

      if (dflag)
         parser->dflag = dflag;

      if (parser->dflag) {
         ArgusLog(LOG_WARNING, "started");
         if (chdir ("/") < 0)
            ArgusLog (LOG_ERR, "Can't chdir to / %s", strerror(errno));

         if ((pid = fork ()) < 0) {
            ArgusLog (LOG_ERR, "Can't fork daemon %s", strerror(errno));
         } else {
            if (pid) {
               int status;
               usleep(20000);
               waitpid(pid, &status, WNOHANG);
               if (kill(pid, 0) < 0) {
                  exit (1);
               } else
                  exit (0);
            } else {
               FILE *tmpfile = NULL;

               parser->ArgusSessionId = setsid();
 
               if ((tmpfile = freopen ("/dev/null", "w", stdout)) == NULL)
                  ArgusLog (LOG_ERR, "Cannot map stdout to /dev/null");
 
               if ((tmpfile = freopen ("/dev/null", "w", stderr)) == NULL)
                  ArgusLog (LOG_ERR, "Cannot map stderr to /dev/null");
            }
         }
      }

      if (parser->ArgusPortNum != 0) 
         if (ArgusEstablishListen (parser, parser->ArgusPortNum, parser->ArgusBindAddr) < 0)
            ArgusLog (LOG_ERR, "setArgusPortNum: ArgusEstablishListen returned %s", strerror(errno));
      
/*
   So the idea here is to create the output queue that will be used
   by RaProcessRecord() routine to queue records for the output
   processor, which will be started when we crank a thread.
   This is the basic new argus() strategy for processing output
   records.  The thread will do two basic things: 
      1) it will grab stuff off the queue, and then do the basic
         processing that this radium will do, such as time
         adjustment, aggregation, correction, and anonymization, etc...

      2) it will manage the listen, to deal without remote client
         requests.  radium() can write its records to a file, and
         any number of remote clients, so ......

   The ArgusClientTimeout() routine will drive all the maintenance
   and so it should be run, probably 4x a second, just for good
   measure.
   
*/
      parser->ArgusOutputList = ArgusNewList();
      parser->ArgusReliableConnection++;
      parser->RaInitialized++;
      parser->ArgusOutput = ArgusNewOutput (parser);
      ArgusInitOutput (parser->ArgusOutput);

      tvp = getArgusMarReportInterval(ArgusParser);
      if ((tvp->tv_sec == 0) && (tvp->tv_usec == 0)) {
         setArgusMarReportInterval (ArgusParser, "60s");
      }
   }
}

void RaArgusInputComplete (struct ArgusInput *input)
{

};

void
RaParseComplete (int sig)
{
   struct ArgusRecordStruct *rec = NULL;
   int n, i;

#if defined(ARGUS_THREADS)
   void *retn = NULL;
#endif

   ArgusParser->RaParseDone++;
#if defined(ARGUS_THREADS)
   if (ArgusParser->remote > 0)
      pthread_join(ArgusParser->remote, &retn);
#endif
   
   if ((n = ArgusParser->ArgusActiveFDs) > 0) {
      for (i = 0; i < n; i++) {  
         if (ArgusParser->ArgusRemoteFDs[i] != NULL) {
            ArgusCloseInput(ArgusParser, ArgusParser->ArgusRemoteFDs[i]);
            ArgusParser->ArgusRemoteFDs[i] = NULL;
         }
      }
   }

   if (ArgusParser->ArgusOutput) {
      if ((rec = ArgusGenerateStatusMarRecord(ArgusParser->ArgusOutput, ARGUS_SHUTDOWN)) != NULL) {
         ArgusPushBackList(ArgusParser->ArgusOutput->ArgusInputList, (struct ArgusListRecord *)rec, ARGUS_LOCK);
      }
   
      ArgusCloseOutput(ArgusParser->ArgusOutput);
   }

   if (sig >= 0) {
      if (sig == SIGINT) {
#if defined(ARGUS_THREADS)
         pthread_exit(0);
#else
         exit(0);
#endif /* ARGUS_THREADS */
      }
   }

#ifdef ARGUSDEBUG
   ArgusDebug (2, "RaParseComplete(%d)\n", sig);
#endif
}


void
ArgusClientTimeout ()
{
   gettimeofday(&ArgusParser->ArgusRealTime, 0);

   if (RaRealTime) {  /* establish value for time comparison */
      if (ArgusLastTime.tv_sec != 0) {
         if (ArgusLastRealTime.tv_sec > 0) {
            dRealTime = *RaDiffTime(&ArgusParser->ArgusRealTime, &ArgusLastRealTime);
            thisUsec = ((dRealTime.tv_sec * 1000000) + dRealTime.tv_usec) * RaUpdateRate;
            dRealTime.tv_sec  = thisUsec / 1000000;
            dRealTime.tv_usec = thisUsec % 1000000;
                                                                                                                             
            ArgusLastTime.tv_sec  += dRealTime.tv_sec;
            ArgusLastTime.tv_usec += dRealTime.tv_usec;
                                                                                                                             
            if (ArgusLastTime.tv_usec > 1000000) {
               ArgusLastTime.tv_sec++;
               ArgusLastTime.tv_usec -= 1000000;
            }
         }

         ArgusLastRealTime = ArgusParser->ArgusRealTime;
      }
   }

   if ((ArgusParser->ArgusReportTime.tv_sec  < ArgusParser->ArgusRealTime.tv_sec) ||
      ((ArgusParser->ArgusReportTime.tv_sec == ArgusParser->ArgusRealTime.tv_sec) &&
       (ArgusParser->ArgusReportTime.tv_usec < ArgusParser->ArgusRealTime.tv_usec))) {

      struct timeval *tvp = getArgusMarReportInterval(ArgusParser);

      if (ArgusParser->ArgusOutput) {
         struct ArgusRecordStruct *rec = NULL;
         if ((rec = ArgusGenerateStatusMarRecord(ArgusParser->ArgusOutput, ARGUS_STATUS)) != NULL)
            ArgusPushBackList(ArgusParser->ArgusOutput->ArgusInputList, (struct ArgusListRecord *)rec, ARGUS_LOCK);
      }

      if (tvp != NULL) {
         ArgusParser->ArgusReportTime.tv_sec  += tvp->tv_sec;
         ArgusParser->ArgusReportTime.tv_usec += tvp->tv_usec;
         if (ArgusParser->ArgusReportTime.tv_usec > 1000000) {
            ArgusParser->ArgusReportTime.tv_sec++;
            ArgusParser->ArgusReportTime.tv_usec -= 1000000;
         }
      }
   }

#ifdef ARGUSDEBUG
   ArgusDebug (6, "ArgusClientTimeout()\n");
#endif
}

void
parse_arg (int argc, char**argv)
{}

void
usage ()
{
   extern char version[];

   fprintf (stderr, "Radium Version %s\n", version);
   fprintf (stderr, "usage: %s [-d] [-f conf] [raoptions]\n", ArgusParser->ArgusProgramName);

   fprintf (stderr, "options: -d             run as a daemon.\n");
   fprintf (stderr, "         -f conf.file   read %s configure file.\n", ArgusParser->ArgusProgramName);
#if defined (ARGUSDEBUG)
   fprintf (stderr, "         -D <level>     specify debug level\n");
#endif
#ifdef ARGUS_SASL
   fprintf (stderr, "         -U <user/auth> specify <user/auth> authentication information.\n");
#endif
   exit(1);
}


void
RaProcessRecord (struct ArgusParserStruct *parser, struct ArgusRecordStruct *argus)
{
   if (argus->hdr.type & ARGUS_MAR) {
      struct ArgusRecord *rec = (struct ArgusRecord *)&argus->canon;
      if (parser->ArgusAdjustTime) {
         struct timeval drift;

         drift.tv_sec  = parser->ArgusRealTime.tv_sec  - ntohl(rec->argus_mar.now.tv_sec);
         drift.tv_usec = parser->ArgusRealTime.tv_usec - ntohl(rec->argus_mar.now.tv_usec);
         parser->ArgusCurrentInput->ArgusTimeDrift  = drift.tv_sec * 1000000;
         parser->ArgusCurrentInput->ArgusTimeDrift += drift.tv_usec;
         rec->argus_mar.drift = parser->ArgusCurrentInput->ArgusTimeDrift;
#ifdef ARGUSDEBUG
#if defined(__APPLE_CC__) || defined(__APPLE__)
         ArgusDebug (3, "RaProcessRecord: ArgusInput 0x%x drift %lld\n", parser->ArgusCurrentInput, parser->ArgusCurrentInput->ArgusTimeDrift);
#else
         ArgusDebug (3, "RaProcessRecord: ArgusInput 0x%x drift %Ld\n", parser->ArgusCurrentInput, parser->ArgusCurrentInput->ArgusTimeDrift);
#endif
#endif
      }

   } else {
      if (parser->ArgusAdjustTime) {
         long long ArgusDriftLevel = parser->ArgusAdjustTime * 1000000;

         if ((parser->ArgusCurrentInput->ArgusTimeDrift >  ArgusDriftLevel) || 
             (parser->ArgusCurrentInput->ArgusTimeDrift < -ArgusDriftLevel)) {
            int secs  = parser->ArgusCurrentInput->ArgusTimeDrift / 1000000;
            int usecs = parser->ArgusCurrentInput->ArgusTimeDrift % 1000000;

            struct timeval startbuf, *start = &startbuf;
            struct timeval endbuf, *end = &endbuf;

            start->tv_sec  = argus->canon.time.src.start.tv_sec;
            start->tv_usec = argus->canon.time.src.start.tv_usec;

            end->tv_sec    = argus->canon.time.src.end.tv_sec;
            end->tv_usec   = argus->canon.time.src.end.tv_usec;

#ifdef ARGUSDEBUG
            ArgusDebug (4, "ArgusHandleDatum() ArgusInput 0x%x adjusting timestamps by %d secs and %d usecs\n", parser->ArgusCurrentInput, secs, usecs);
#endif
            argus->canon.time.hdr.argus_dsrvl8.qual |= ARGUS_TIMEADJUST;
            start->tv_sec  +=  secs;
            start->tv_usec += usecs;
            if (start->tv_usec < 0) {
               start->tv_sec--; start->tv_usec += 1000000;
            }
            if (start->tv_usec > 1000000) {
               start->tv_sec++; start->tv_usec -= 1000000;
            }

            end->tv_sec  +=  secs;
            end->tv_usec += usecs;
            if (end->tv_usec < 0) {
               end->tv_sec--; end->tv_usec += 1000000;
            }
            if (end->tv_usec > 1000000) {
               end->tv_sec++; end->tv_usec -= 1000000;
            }
         }
      }

      ArgusThisTime.tv_sec  = argus->canon.time.src.start.tv_sec;
      ArgusThisTime.tv_usec = argus->canon.time.src.start.tv_usec;

      if (RaRealTime) {
         if (ArgusLastTime.tv_sec == 0)
            ArgusLastTime = ArgusThisTime;
                                                                                                                           
         if (!((ArgusLastTime.tv_sec  > ArgusThisTime.tv_sec) ||
            ((ArgusLastTime.tv_sec == ArgusThisTime.tv_sec) &&
             (ArgusLastTime.tv_usec > ArgusThisTime.tv_usec)))) {
                                                                                                                           
            while ((ArgusThisTime.tv_sec  > ArgusLastTime.tv_sec) ||
                  ((ArgusThisTime.tv_sec == ArgusLastTime.tv_sec) &&
                   (ArgusThisTime.tv_usec > ArgusLastTime.tv_usec))) {
               int thisRate;
                                                                                                                           
               dThisTime = *RaDiffTime(&ArgusThisTime, &ArgusLastTime);
               thisRate = ((dThisTime.tv_sec * 1000000) + dThisTime.tv_usec)/RaUpdateRate;
               thisRate = (thisRate > 100000) ? 100000 : thisRate;

               usleep (thisRate);
               ArgusClientTimeout ();

               gettimeofday(&parser->ArgusRealTime, 0);

               if (ArgusLastRealTime.tv_sec > 0) {
                  dRealTime = *RaDiffTime(&parser->ArgusRealTime, &ArgusLastRealTime);
                  thisUsec = ((dRealTime.tv_sec * 1000000) + dRealTime.tv_usec) * RaUpdateRate;
                  dRealTime.tv_sec  = thisUsec / 1000000;
                  dRealTime.tv_usec = thisUsec % 1000000;
                                                                                                                           
                  ArgusLastTime.tv_sec  += dRealTime.tv_sec;
                  ArgusLastTime.tv_usec += dRealTime.tv_usec;
                  if (ArgusLastTime.tv_usec > 1000000) {
                     ArgusLastTime.tv_sec++;
                     ArgusLastTime.tv_usec -= 1000000;
                  }
               }
                                                                                                                           
               ArgusLastRealTime = parser->ArgusRealTime;
            }
         }
                                                                                                                           
      } else
         ArgusLastTime = parser->ArgusRealTime;
   }

   ArgusPushBackList(parser->ArgusOutputList, (struct ArgusListRecord *)ArgusCopyRecordStruct(argus), ARGUS_LOCK);

#ifdef ARGUSDEBUG
   ArgusDebug (4, "RaProcessRecord (0x%x, 0x%x) returning", parser, argus); 
#endif
}

int RaSendArgusRecord(struct ArgusRecordStruct *argus) {return 0;}

void ArgusWindowClose(void);

void ArgusWindowClose(void) { 
#ifdef ARGUSDEBUG
   ArgusDebug (6, "ArgusWindowClose () returning\n"); 
#endif
}


#define RADIUM_RCITEMS				18

#define RADIUM_DAEMON				0
#define RADIUM_MONITOR_ID			1
#define RADIUM_ARGUS_SERVER			2
#define RADIUM_CISCONETFLOW_PORT                3
#define RADIUM_ACCESS_PORT			4
#define RADIUM_INPUT_FILE                       5
#define RADIUM_USER_AUTH			6 
#define RADIUM_AUTH_PASS			7
#define RADIUM_OUTPUT_FILE			8
#define RADIUM_MAR_STATUS_INTERVAL		9
#define RADIUM_DEBUG_LEVEL			10
#define RADIUM_FILTER_OPTIMIZER			11
#define RADIUM_FILTER_TAG			12
#define RADIUM_BIND_IP				13
#define RADIUM_MIN_SSF				14
#define RADIUM_MAX_SSF				15
#define RADIUM_ADJUST_TIME			16
#define RADIUM_ARCHIVE				17

char *RadiumResourceFileStr [] = {
   "RADIUM_DAEMON=",
   "RADIUM_MONITOR_ID=",
   "RADIUM_ARGUS_SERVER=",
   "RADIUM_CISCONETFLOW_PORT=",
   "RADIUM_ACCESS_PORT=",
   "RADIUM_INPUT_FILE=",
   "RADIUM_USER_AUTH=",
   "RADIUM_AUTH_PASS=",
   "RADIUM_OUTPUT_FILE=",
   "RADIUM_MAR_STATUS_INTERVAL=",
   "RADIUM_DEBUG_LEVEL=",
   "RADIUM_FILTER_OPTIMIZER=",
   "RADIUM_FILTER=",
   "RADIUM_BIND_IP=",
   "RADIUM_MIN_SSF=",
   "RADIUM_MAX_SSF=",
   "RADIUM_ADJUST_TIME=",
   "RADIUM_ARCHIVE=",
};


int
RadiumParseResourceFile (struct ArgusParserStruct *parser, char *file)
{
   int retn = 0;
   int i, len, done = 0, linenum = 0, Soption = 0, roption = 0;
   char strbuf[MAXSTRLEN], *str = strbuf, *optarg;
   char result[MAXSTRLEN], *ptr;
   struct hostent *host;
   FILE *fd;

   if (file) {
      if ((fd = fopen (file, "r")) != NULL) {
         while ((fgets(str, MAXSTRLEN, fd)) != NULL)  {
            done = 0;
            linenum++;
            while (*str && isspace((int)*str))
                str++;

            if (*str && (*str != '#') && (*str != '\n') && (*str != '!')) {
               for (i = 0; i < RADIUM_RCITEMS && !done; i++) {
                  len = strlen(RadiumResourceFileStr[i]);
                  if (!(strncmp (str, RadiumResourceFileStr[i], len))) {
                     optarg = &str[len];
                     if (*optarg == '\"') optarg++;
                     if (optarg[strlen(optarg) - 1] == '\n')
                        optarg[strlen(optarg) - 1] = '\0';
                     if (optarg[strlen(optarg) - 1] == '\"')
                        optarg[strlen(optarg) - 1] = '\0';

                     switch (i) {
                        case RADIUM_DAEMON: 
                           if (!(strncasecmp(optarg, "yes", 3)))
                              parser->dflag = 1;
                           else
                           if (!(strncasecmp(optarg, "no", 2)))
                              parser->dflag = 0;
                           break;

                        case RADIUM_MONITOR_ID: 
                           if (optarg && (*optarg == '`')) {
                              if (optarg[strlen(optarg) - 1] == '`') {
                                 FILE *tfd;

                                 optarg++;
                                 optarg[strlen(optarg) - 1] = '\0';
                                 if (!(strcmp (optarg, "hostname"))) {
                                    if ((tfd = popen("hostname", "r")) != NULL) {
                                       if ((ptr = fgets(result, MAXSTRLEN, tfd)) != NULL) {
                                          optarg = ptr;
                                          optarg[strlen(optarg) - 1] = '\0';
                                       } else
                                          ArgusLog (LOG_ERR, "RadiumParseResourceFile(%s) `hostname` failed %s.\n", file, strerror(errno));

                                       pclose(tfd);
                                    } else
                                       ArgusLog (LOG_ERR, "RadiumParseResourceFile(%s) System error: popen() %s\n", file, strerror(errno));
                                 } else
                                    ArgusLog (LOG_ERR, "RadiumParseResourceFile(%s) unsupported command `%s` at line %d.\n", file, optarg, linenum);
                              } else
                                 ArgusLog (LOG_ERR, "RadiumParseResourceFile(%s) syntax error line %d\n", file, linenum);
                           }
                           if (optarg && isalnum((int)*optarg)) {
                              if ((host = gethostbyname(optarg)) != NULL) {
                                 if ((host->h_addrtype == 2) && (host->h_length == 4)) {
                                    unsigned int addr;
                                    bcopy ((char *) *host->h_addr_list, (char *)&addr, host->h_length);
                                    setArgusID (parser, ntohl(addr));
                                 } else
                                    ArgusLog (LOG_ERR, "RadiumParseResourceFile(%s) host '%s' error %s\n", file, optarg, strerror(errno));
                                 setArgusIDType(parser, ARGUS_ID_IS_IPADDR);
                              } else
                                 if (optarg && isdigit((int)*optarg)) {
                                    setArgusID (parser, atoi (optarg));
                                 } else
                                    ArgusLog (LOG_ERR, "RadiumParseResourceFile(%s) syntax error line %d\n", file, linenum);
                           } else
                              ArgusLog (LOG_ERR, "RadiumParseResourceFile(%s) syntax error line %d\n", file, linenum);

                           break;

                        case RADIUM_ARGUS_SERVER:
                           ++parser->Sflag;
                           if (!Soption++ && (parser->ArgusRemoteHostList != NULL))
                              ArgusDeleteHostList(parser);

                           if (!(ArgusAddHostList (parser, optarg, ARGUS_DATA_SOURCE)))
                              ArgusLog (LOG_ERR, "%s: host %s unknown\n", optarg);
                           break;

                        case RADIUM_CISCONETFLOW_PORT: {
                           ++parser->Sflag; ++parser->Cflag;
                           if (!Soption++ && (parser->ArgusRemoteHostList != NULL))
                              ArgusDeleteHostList(parser);

                           if (!(ArgusAddHostList (parser, optarg, ARGUS_CISCO_DATA_SOURCE)))
                              ArgusLog (LOG_ERR, "%s: host %s unknown\n", optarg);

                           break;
                        }

                        case RADIUM_INPUT_FILE:
                           if ((!roption++) && (parser->ArgusInputFileList != NULL))
                              ArgusDeleteFileList(parser);

                           if (!(ArgusAddFileList (parser, optarg, (parser->Cflag ? ARGUS_CISCO_DATA_SOURCE : ARGUS_DATA_SOURCE), -1, -1))) {
                              ArgusLog (LOG_ERR, "%s: error: file arg %s\n", optarg);
                           }
                           break;
                           
                        case RADIUM_ACCESS_PORT:
                           parser->ArgusPortNum = atoi(optarg);
                           break;
/*
                        case RADIUM_USER_AUTH:
                           ustr = strdup(optarg);
                           break;

                        case RADIUM_AUTH_PASS:
                           pstr = strdup(optarg);
                           break;
*/
                        case RADIUM_OUTPUT_FILE:
                           ptr = NULL;
                           if ((ptr = strchr (optarg, '"')) != NULL)
                              *ptr = '\0';

                           if ((ptr = strchr (optarg, ' ')) != NULL) {
                              *ptr++ = '\0';
                           }
                   
                           setArgusWfile (parser, optarg, ptr);
                           break;

                        case RADIUM_MAR_STATUS_INTERVAL:
                           setArgusMarReportInterval (parser, optarg);
                           break;

                        case RADIUM_DEBUG_LEVEL:
                           parser->debugflag = atoi(optarg);
                           break;
                        
                        case RADIUM_FILTER_OPTIMIZER:
                           if ((strncasecmp(optarg, "yes", 3)))
                              setArgusOflag  (parser, 1);
                           else
                              setArgusOflag  (parser, 0);
                           break;

                        case RADIUM_FILTER_TAG:
                           if ((parser->ArgusRemoteFilter = ArgusCalloc (1, MAXSTRLEN)) != NULL) {
                              ptr = parser->ArgusRemoteFilter;
                              str = optarg;
                              while (*str) {
                                 if ((*str == '\\') && (str[1] == '\n')) {
                                    fgets(str, MAXSTRLEN, fd);
                                    while (*str && (isspace((int)*str) && (str[1] && isspace((int)str[1]))))
                                       str++;
                                 }
                                 
                                 if ((*str != '\n') && (*str != '"'))
                                    *ptr++ = *str++;
                                 else
                                    str++;
                              }
#ifdef ARGUSDEBUG
                              ArgusDebug (1, "RadiumParseResourceFile: ArgusFilter \"%s\" \n", parser->ArgusRemoteFilter);
#endif 
                           }
                           break;

                        case RADIUM_BIND_IP:
                           if (*optarg != '\0')
                              setArgusBindAddr (parser, optarg);
#ifdef ARGUSDEBUG
                           ArgusDebug (1, "RadiumParseResourceFile: ArgusBindAddr \"%s\" \n", parser->ArgusBindAddr);
#endif 
                           break;

                        case RADIUM_MIN_SSF:
                           if (*optarg != '\0') {
#ifdef ARGUS_SASL
                              ArgusMinSsf = atoi(optarg);
#ifdef ARGUSDEBUG
                           ArgusDebug (1, "RadiumParseResourceFile: ArgusMinSsf \"%s\" \n", ArgusMinSsf);
#endif
#endif
                           }
                           break;

                        case RADIUM_MAX_SSF:
                           if (*optarg != '\0') {
#ifdef ARGUS_SASL
                              ArgusMaxSsf = atoi(optarg);
#ifdef ARGUSDEBUG
                              ArgusDebug (1, "RadiumParseResourceFile: ArgusMaxSsf \"%s\" \n", ArgusMaxSsf);
#endif
#endif
                           }
                           break;

                        case RADIUM_ADJUST_TIME: {
                           char *ptr;
                           parser->ArgusAdjustTime = strtol(optarg, (char **)&ptr, 10);
                           if (ptr == optarg)
                              ArgusLog (LOG_ERR, "%s syntax error: line %d", file, linenum);
                           
                           if (isalpha((int) *ptr)) {
                              switch (*ptr) {
                                 case 's': break;
                                 case 'm': parser->ArgusAdjustTime *= 60; break;
                                 case 'h': parser->ArgusAdjustTime *= 3600; break;
                              }
                           }
#ifdef ARGUSDEBUG
                           ArgusDebug (1, "RadiumParseResourceFile: ArgusAdjustTime is %d secs\n", parser->ArgusAdjustTime);
#endif
                           break;
                        }

                        case RADIUM_ARCHIVE: {
                           setArgusArchive (parser, optarg);
#ifdef ARGUSDEBUG
                           ArgusDebug (1, "RadiumParseResourceFile: Radium Archive is %s\n", optarg);
#endif
                        }
                     }

                     done = 1;
                     break;
                  }
               }
            }
         }

         fclose(fd);

      } else {
         retn++;
#ifdef ARGUSDEBUG
         ArgusDebug (1, "RadiumParseResourceFile: open %s %s\n", file, strerror(errno));
#endif 
      }
   }

#ifdef ARGUSDEBUG
   ArgusDebug (1, "RadiumParseResourceFile (%s) returning %d\n", file, retn);
#endif 

   return (retn);
}

void
clearRadiumConfiguration (void)
{
   ArgusParser->dflag = 0;
   setArgusID (ArgusParser, 0);
   setArgusIDType (ArgusParser, 0);

   ArgusParser->ArgusPortNum = 0;

   clearArgusWfile (ArgusParser);
   setArgusBindAddr (ArgusParser, NULL);
   setArgusOflag (ArgusParser, 1);

   ArgusParser->dflag = 0;

   if (ArgusParser->ArgusRemoteHostList != NULL)
      ArgusDeleteHostList(ArgusParser);

   if (ArgusParser->ArgusInputFileList) {
      ArgusDeleteFileList(ArgusParser);
   }
 
   if (ArgusParser->ArgusRemoteFilter) {
      ArgusFree(ArgusParser->ArgusRemoteFilter);
      ArgusParser->ArgusRemoteFilter = NULL;
   }

#ifdef ARGUSDEBUG
   ArgusDebug (1, "clearRadiumConfiguration () returning\n");
#endif 
}


void
setArgusPortNum (struct ArgusParserStruct *parser, int value, char *addr)
{
   parser->ArgusPortNum = value;

#ifdef ARGUSDEBUG
   ArgusDebug (2, "setArgusPortNum(%d) returning\n", value);
#endif
}

int
getArgusPortNum(struct ArgusParserStruct *parser)
{
   return(parser->ArgusPortNum);
}


void
setArgusID(struct ArgusParserStruct *parser, unsigned int value)
{
   parser->ArgusID = value;
}

unsigned int
getArgusID(struct ArgusParserStruct *parser) {
   return (parser->ArgusID);
}


void
setArgusOflag(struct ArgusParserStruct *parser, unsigned int value)
{
   parser->Oflag = value;
}

void
setArgusIDType(struct ArgusParserStruct *parser, unsigned int value)
{
   parser->ArgusIDType = value;
}

unsigned int
getArgusIDType(struct ArgusParserStruct *parser)
{
   return (parser->ArgusIDType);
}



#include <ctype.h>
#include <math.h>

void
setArgusMarReportInterval (struct ArgusParserStruct *parser, char *value)
{
   struct timeval *tvp = getArgusMarReportInterval(parser);

   struct timeval ovalue;
   double thisvalue = 0.0, iptr, fptr;
   int ivalue = 0;
   char *ptr = NULL;

   gettimeofday (&parser->ArgusRealTime, 0L);

   if (tvp != NULL) {
      ovalue = *tvp;
      tvp->tv_sec  = 0;
      tvp->tv_usec = 0;
   } else {
      ovalue.tv_sec  = 0;
      ovalue.tv_usec = 0;
   }

   if (((ptr = strchr (value, '.')) != NULL) || isdigit((int)*value)) {
      if (ptr != NULL) {
         thisvalue = atof(value);
      } else {
         if (isdigit((int)*value)) {
            ivalue = atoi(value);
            thisvalue = ivalue * 1.0;
         }
      }

      fptr =  modf(thisvalue, &iptr);

      tvp->tv_sec = iptr;
      tvp->tv_usec =  fptr * 1000000;

      parser->ArgusReportTime.tv_sec  = parser->ArgusRealTime.tv_sec + tvp->tv_sec;
      parser->ArgusReportTime.tv_usec = tvp->tv_usec;

   } else
      *tvp = ovalue;

#ifdef ARGUSDEBUG
   ArgusDebug (4, "setArgusMarReportInterval(%s) returning\n", value);
#endif
}

void
setArgusBindAddr (struct ArgusParserStruct *parser, char *value)
{
   if (parser->ArgusBindAddr != NULL)
      free(parser->ArgusBindAddr);

   if (value != NULL)
      parser->ArgusBindAddr = strdup(value);
   else
      parser->ArgusBindAddr = NULL;
}


struct timeval *
getArgusMarReportInterval(struct ArgusParserStruct *parser) {
   return (&parser->ArgusMarReportInterval);
}


#include <netdb.h>

int
ArgusEstablishListen (struct ArgusParserStruct *parser, int port, char *bindIP)
{
   int s = -1;
   struct sockaddr_in sin;
   struct timeval tvpbuf, *tvp = &tvpbuf;
   struct hostent *host;

   gettimeofday (tvp, 0L);

   if (port) {
      sin.sin_addr.s_addr = INADDR_ANY;
      if (bindIP) {
#ifdef ARGUSDEBUG
         ArgusDebug (1, "ArgusEstablishListen(%d, %s, 0x%x)\n", port, bindIP);
#endif
         if ((host = gethostbyname (bindIP)) != NULL) {
            if ((host->h_addrtype == AF_INET) && (host->h_length == 4)) {
               bcopy ((char *) *host->h_addr_list, (char *)&sin.sin_addr.s_addr, host->h_length);
            } else
               ArgusLog (LOG_ERR, "ArgusEstablishListen() unsupported bind address %s", bindIP);
         } else
            ArgusLog (LOG_ERR, "ArgusEstablishListen() bind address %s error %s", bindIP, strerror(errno));
      }

      sin.sin_port = htons((u_short) port);
      sin.sin_family = AF_INET;

#ifdef ARGUSDEBUG
      ArgusDebug (1, "ArgusEstablishListen(0x%x, %d, %s) binding: %d\n", parser, port, bindIP, sin.sin_addr.s_addr);
#endif

      if ((s = socket (AF_INET, SOCK_STREAM, 0)) != -1) {
         int flags = fcntl (s, F_GETFL, 0L);
         if ((fcntl (s, F_SETFL, flags | O_NDELAY)) >= 0) {
            if (!(bind (s, (struct sockaddr *)&sin, sizeof(sin)))) {
               if ((listen (s, parser->ArgusMaxListen)) >= 0) {
                  parser->ArgusLfd = s;
               } else {
                  close (s);
                  s = -1;
                  ArgusLog(LOG_ERR, "ArgusEstablishListen: listen() failure");
               }
            } else {
               close (s);
               s = -1;
               ArgusLog(LOG_ERR, "ArgusEstablishListen: bind() error");
            }
         } else
            ArgusLog(LOG_ERR, "ArgusEstablishListen: fcntl() error");
      } else
         ArgusLog(LOG_ERR, " ArgusEstablishListen: socket() error");
   }
     
#ifdef ARGUSDEBUG
   ArgusDebug (2, "ArgusEstablishListen(0x%x, %d, 0x%x) returning %d\n", parser, port, bindIP, s);
#endif

   return (s);
}


#ifndef ArgusOutput
#define ArgusOutput
#endif

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
 

#if defined(ARGUS_THREADS)
#include <pthread.h>
#endif


 
#define ARGUS_READINGPREHDR     1
#define ARGUS_READINGHDR        2
#define ARGUS_READINGBLOCK      4
 
#define ARGUS_MAXRECORDS        1024
 
 
#define ARGUS_WAS_FUNCTIONAL             0x10
#define ARGUS_SOCKET_COMPLETE            0x20
#define ARGUS_MAXRECORD                  0x10000
 
struct ArgusSocketStruct {
   struct ArgusListStruct *ArgusOutputList;
   int fd, status, cnt, expectedSize, errornum;
   int ArgusLastRecord, ArgusReadState;
   struct timeval lastwrite;
   unsigned char buf[ARGUS_MAXRECORD], *ptr;
   struct ArgusRecordStruct *rec;
   int length, writen;
   struct sockaddr sock;
   char *filename;
   void *obj;
}; 
 
typedef int (*ArgusHandler)(struct ArgusSocketStruct *, unsigned char *, int, void *);


struct ArgusOutputStruct *
ArgusNewOutput (struct ArgusParserStruct *parser)
{
   struct ArgusOutputStruct *retn = NULL;
   int i;

   if ((retn = (struct ArgusOutputStruct *) ArgusCalloc (1, sizeof (struct ArgusOutputStruct))) == NULL)
     ArgusLog (LOG_ERR, "ArgusNewOutput() ArgusCalloc error %s\n", strerror(errno));

   for (i = 0; i < ARGUS_MAXLISTEN; i++)
      retn->clients[i].fd = -1;

   retn->ArgusPortNum   = parser->ArgusPortNum;
   retn->ArgusBindAddr  = parser->ArgusBindAddr;
   retn->ArgusWfileList = parser->ArgusWfileList;
   retn->ArgusLfd       = parser->ArgusLfd;

   retn->ArgusInputList   = parser->ArgusOutputList;
   parser->ArgusWfileList = NULL;

   retn->ArgusMarReportInterval   = parser->ArgusMarReportInterval;
   retn->ArgusReportTime.tv_sec   = parser->ArgusGlobalTime.tv_sec + parser->ArgusMarReportInterval.tv_sec;
   retn->ArgusReportTime.tv_usec += parser->ArgusMarReportInterval.tv_usec;
   retn->ArgusLastMarUpdateTime   = parser->ArgusGlobalTime;

   gettimeofday (&retn->ArgusStartTime, 0L);

#ifdef ARGUSDEBUG
   ArgusDebug (1, "ArgusNewOutput() returning retn 0x%x\n", retn);
#endif

   return (retn);
}



void
ArgusInitOutput (struct ArgusOutputStruct *output)
{
   struct ArgusWfileStruct *wfile;
   int len = 0;
#if defined(ARGUS_THREADS)
   pthread_attr_t attrbuf, *attr = &attrbuf;
#endif /* ARGUS_THREADS */

   if ((output->ArgusOutputList = ArgusNewList()) == NULL)
      ArgusLog (LOG_ERR, "ArgusInitOutput: ArgusList %s", strerror(errno));

   if (output->ArgusInitMar != NULL)
      ArgusFree (output->ArgusInitMar);

   if ((output->ArgusInitMar = ArgusGenerateInitialMar(output)) == NULL)
      ArgusLog (LOG_ERR, "ArgusInitOutput: ArgusGenerateInitialMar error %s", strerror(errno));

   len = ntohs(output->ArgusInitMar->hdr.len) * 4;

   if (output->ArgusWfileList != NULL) {
      setuid(getuid());

      while (!ArgusListEmpty(output->ArgusWfileList)) {
         if ((wfile = (struct ArgusWfileStruct *) ArgusPopFrontList(output->ArgusWfileList, ARGUS_LOCK)) != NULL) {
            int ind = output->clientindex++;

            if (strcmp (wfile->filename, "-")) {
               if ((output->clients[ind].fd = open (wfile->filename, O_WRONLY|O_APPEND|O_CREAT|O_NONBLOCK, 0x1a4)) < 0)
                  ArgusLog (LOG_ERR, "ArgusInitOutput: open %s: %s", wfile->filename, strerror(errno));
            } else {
               output->clients[ind].fd = 1;
               output->ArgusWriteStdOut++;
            }

            output->clienttags |= (1 << ind);
            output->clientnum++;

            if ((output->clients[ind].sock = ArgusNewSocket(output->clients[ind].fd)) == NULL)
               ArgusLog (LOG_ERR, "ArgusInitOutput: ArgusNewSocket error %s", strerror(errno));

            if (write (output->clients[ind].fd, (char *) output->ArgusInitMar, len) != len) {
               if (!output->ArgusWriteStdOut) {
                  close (output->clients[ind].fd);
                  unlink (wfile->filename);
               }
               ArgusLog (LOG_ERR, "ArgusInitOutput: write(): %s", strerror(errno));
            }

            if (strcmp(wfile->filename, "/dev/null"))
               output->clients[ind].sock->filename = wfile->filename;

            output->clients[ind].ArgusClientStart++;
         }
      }

      ArgusDeleteList(output->ArgusWfileList);
      output->ArgusWfileList = NULL;
   }


#if defined(ARGUS_THREADS)
   pthread_attr_init(attr); 

   if (getuid == 0)
      pthread_attr_setschedpolicy(attr, SCHED_RR);
   else
      attr = NULL;

   if ((pthread_create(&output->thread, attr, ArgusOutputProcess, (void *) output)) != 0)
      ArgusLog (LOG_ERR, "ArgusNewOutput() pthread_create error %s\n", strerror(errno));

#endif /* ARGUS_THREADS */

#ifdef ARGUSDEBUG
   ArgusDebug (1, "ArgusInitOutput() done\n");
#endif
}


void
ArgusCloseOutput(struct ArgusOutputStruct *output)
{
#if defined(ARGUS_THREADS)
   void *retn = NULL;

#ifdef ARGUSDEBUG
   if (output->ArgusOutputList != NULL)
      ArgusDebug (1, "ArgusCloseOutput() scheduling closure after %d records", output->ArgusInputList->count + output->ArgusOutputList->count);
   else
      ArgusDebug (1, "ArgusCloseOutput() closing", output->ArgusOutputList->count);
#endif

   output->status |= ARGUS_SHUTDOWN;
   if (output != NULL) 
      pthread_join(output->thread, &retn);
#else
   if (output != NULL) {
      output->status |= ARGUS_SHUTDOWN;
      ArgusOutputProcess(output);
   }
#endif /* ARGUS_THREADS */

#ifdef ARGUSDEBUG
   ArgusDebug (1, "ArgusCloseOutput() done\n");
#endif
}


void ArgusCheckClientStatus (struct ArgusOutputStruct *);
int ArgusCheckClientMessage (struct ArgusOutputStruct *, int);
int ArgusCongested = 0;

int ArgusCheckClients(struct ArgusOutputStruct *);

int 
ArgusCheckClients(struct ArgusOutputStruct *output)
{
   struct timeval ArgusUpDate = {0, 250000};
   int i, val, retn = 0;

   gettimeofday (&output->ArgusGlobalTime, 0L);

    /* check to see if there are any new clients */

   if ((output->ArgusGlobalTime.tv_sec >  output->ArgusNextUpdate.tv_sec) ||
      ((output->ArgusGlobalTime.tv_sec == output->ArgusNextUpdate.tv_sec) &&
       (output->ArgusGlobalTime.tv_usec > output->ArgusNextUpdate.tv_usec))) {
   
      retn++;
      if (output->ArgusLfd != -1) {
         struct timeval wait = {0, 0}; 
         fd_set readmask;
         int width = output->ArgusLfd;
 
         FD_ZERO(&readmask);
         FD_SET(output->ArgusLfd, &readmask);
         if (output->clienttags) {
            for (i = 0; i < ARGUS_MAXLISTEN; i++) {
               if ((output->clients[i].fd != -1) && (!(output->clients[i].sock->filename))) {
                  if (!(output->clients[i].pid)) {
                     FD_SET(output->clients[i].fd, &readmask);
                     width = (output->clients[i].fd > width) ? output->clients[i].fd : width;
                  }
               }
            }
         }

         if ((val = select (width + 1, &readmask, NULL, NULL, &wait)) >= 0) {
            if (val > 0) {
               if (FD_ISSET(output->ArgusLfd, &readmask)) {
                  ArgusCheckClientStatus(output);

               } else {
#ifdef ARGUSDEBUG
                  ArgusDebug (3, "ArgusCheckClient() select returned with read request\n");
#endif
                  for (i = 0; i < ARGUS_MAXLISTEN; i++) {
                     if (output->clients[i].fd != -1) {
                        if (FD_ISSET(output->clients[i].fd, &readmask)) {
                           if (ArgusCheckClientMessage(output, i) < 0) {
                              if (output->clients[i].pid == 0) {
                                 ArgusDeleteSocket(output, output->clients[i].sock);
                                 output->clienttags &= ~(1 << i);
                                 output->clientnum--;
                                 output->clients[i].fd = -1;
                              }
                           }
                        }
                     }
                  }
               }
            }
         }

         output->ArgusNextUpdate.tv_usec += ArgusUpDate.tv_usec;
         output->ArgusNextUpdate.tv_sec  += ArgusUpDate.tv_sec;

         if (output->ArgusNextUpdate.tv_usec > 1000000) {
            output->ArgusNextUpdate.tv_sec++;
            output->ArgusNextUpdate.tv_usec -= 1000000;
         }

#ifdef ARGUSDEBUG
         ArgusDebug (9, "ArgusCheckClients(struct ArgusOutputStruct 0x%x) returns %d\n", output, retn);
#endif
      }
   }

   return (retn);
}

void *
ArgusOutputProcess(void *arg)
{
   struct ArgusOutputStruct *output = (struct ArgusOutputStruct *) arg;
   struct ArgusRecordStruct *rec = NULL;
   void *retn = NULL;
   int i, count, pid;

#if defined(ARGUS_THREADS)
   sigset_t blocked_signals;

   sigemptyset(&blocked_signals);
   sigaddset(&blocked_signals, SIGTERM);
   sigaddset(&blocked_signals, SIGKILL);
   sigaddset(&blocked_signals, SIGINT);
   sigaddset(&blocked_signals, SIGSTOP);
   sigaddset(&blocked_signals, SIGUSR1);
   sigaddset(&blocked_signals, SIGUSR2);
   sigaddset(&blocked_signals, SIGALRM);
   sigaddset(&blocked_signals, SIGPIPE);
   pthread_sigmask(SIG_BLOCK, &blocked_signals, NULL);
#endif

#ifdef ARGUSDEBUG
   ArgusDebug (8, "ArgusOutputProcess(0x%x) starting\n", output);
#endif

#if defined(ARGUS_THREADS)
   while (!((output->status & ARGUS_STOP) || (output->status & ARGUS_SHUTDOWN))) {
#else
   {
#endif

      if (output && output->ArgusInputList) {
         ArgusCheckClients(output);

#if defined(ARGUS_THREADS)
         if (ArgusListEmpty(output->ArgusInputList))
            usleep(10000);
#endif
         if (!(ArgusListEmpty(output->ArgusInputList))) {
            ArgusLoadList(output->ArgusInputList, output->ArgusOutputList);

            while ((rec = (struct ArgusRecordStruct *) ArgusPopFrontList(output->ArgusOutputList, ARGUS_LOCK)) != NULL) {
               output->ArgusTotalRecords++;

               if (((rec->hdr.type & 0xF0) == ARGUS_MAR) && 
                  (((rec->hdr.cause & 0xF0) == ARGUS_STOP) || ((rec->hdr.cause & 0xF0) == ARGUS_SHUTDOWN))) {
                  if (((struct ArgusRecord *)&rec->canon)->argus_mar.argusid == getArgusID(ArgusParser)) {
                     output->status |= ARGUS_STOP;
#ifdef ARGUSDEBUG
                     ArgusDebug (2, "ArgusOutputProcess() received stop record %d records on the list\n", output->ArgusOutputList->count);
#endif
                  }
               }

               if ((count = output->clientnum) > 0) {
                  for (i = 0; i < ARGUS_MAXLISTEN; i++) {
                     if ((output->clients[i].fd != -1) && (output->clients[i].sock != NULL) && output->clients[i].ArgusClientStart) {
                        int arguswriterecord = 1;

                        if (output->clients[i].ArgusFilterInitialized) {
                           if (!(ArgusFilterRecord ((struct nff_insn *)output->clients[i].ArgusBPFcode.bf_insns, rec)))
                              arguswriterecord = 0;
                        }

                        if (arguswriterecord) {
                           if (ArgusWriteSocket (output, output->clients[i].sock, rec, count--) < 0) {
                              ArgusDeleteSocket(output, output->clients[i].sock);
                              output->clienttags &= ~(1 << i);
                              output->clientnum--;
                              output->clients[i].fd = -1;

                           } else {
                              if (ArgusWriteOutSocket (output, output->clients[i].sock) < 0) {
                                 ArgusDeleteSocket(output, output->clients[i].sock);
                                 output->clienttags &= ~(1 << i);
                                 output->clientnum--;
                                 output->clients[i].fd = -1;
                              }
                           }
                        } else {
                           if (--count == 0)
                              ArgusFreeListRecord(rec);
                        }
                     }
                  }

               } else
                  ArgusFreeListRecord(rec);
            }

            if (output->ArgusWriteStdOut)
               fflush (stdout);
         }

         if (output->clienttags) {
            for (i = 0; i < ARGUS_MAXLISTEN; i++) {
               if ((pid = output->clients[i].pid) != 0) {
                  int status;
                  if ((pid = waitpid(pid, &status, WNOHANG)) == output->clients[i].pid) {
#ifdef ARGUSDEBUG
                     ArgusDebug (2, "ArgusOutputProcess(0x%x) file processor %d done\n", output, pid);
#endif
                     ArgusDeleteSocket(output, output->clients[i].sock);
                     output->clienttags &= ~(1 << i);
                     output->clientnum--;
                     output->clients[i].fd  = -1;
                     output->clients[i].pid = 0;
                  }

               } else {
                  if ((output->clients[i].fd != -1) && (output->clients[i].sock != NULL)) {
                     if (ArgusWriteOutSocket (output, output->clients[i].sock) < 0) {
                        ArgusDeleteSocket(output, output->clients[i].sock);
                        output->clienttags &= ~(1 << i);
                        output->clientnum--;
                        output->clients[i].fd = -1;
                     }
                  }
               }
            }
         }

      } else {
#ifdef ARGUSDEBUG
         ArgusDebug (2, "ArgusOutputProcess() waiting for ArgusOutputList 0x%x\n", output);
#endif
         usleep (25000);
      }
#if !defined(ARGUS_THREADS)
   }
#else
   }
#endif /* ARGUS_THREADS */


   if (!(ArgusListEmpty(output->ArgusInputList))) {
      ArgusLoadList(output->ArgusInputList, output->ArgusOutputList);

      while ((rec = (struct ArgusRecordStruct *) ArgusPopFrontList(output->ArgusOutputList, ARGUS_LOCK)) != NULL) {
         if (((rec->hdr.type & 0xF0) == ARGUS_MAR) && ((rec->hdr.cause & 0xF0) == ARGUS_STOP)) {
            if (((struct ArgusRecord *)&rec->canon)->argus_mar.argusid == getArgusID(ArgusParser))
               output->status |= ARGUS_STOP;
#ifdef ARGUSDEBUG
            ArgusDebug (2, "ArgusOutputProcess() received stop record %d records on the list\n", output->ArgusOutputList->count);
#endif
         }

         if ((count = output->clientnum) > 0) {
            for (i = 0; i < ARGUS_MAXLISTEN; i++) {
               if ((output->clients[i].fd != -1) && (output->clients[i].sock != NULL) && output->clients[i].ArgusClientStart) {
                  int arguswriterecord = 1;

                  if (output->clients[i].ArgusFilterInitialized) {
                     if (!(ArgusFilterRecord ((struct nff_insn *)output->clients[i].ArgusBPFcode.bf_insns, rec)))
                        arguswriterecord = 0;
                  }

                  if (arguswriterecord) {
                     if (ArgusWriteSocket (output, output->clients[i].sock, rec, count--) < 0) {
                        ArgusDeleteSocket(output, output->clients[i].sock);
                        output->clienttags &= ~(1 << i);
                        output->clientnum--;
                        output->clients[i].fd = -1;

                     } else {
                        if (ArgusWriteOutSocket (output, output->clients[i].sock) < 0) {
                           ArgusDeleteSocket(output, output->clients[i].sock);
                           output->clienttags &= ~(1 << i);
                           output->clientnum--;
                           output->clients[i].fd = -1;
                        }
                     }
                  }
               }
            }
         } else
            ArgusFreeListRecord(rec);
      }

      if (output->ArgusWriteStdOut)
         fflush (stdout);
   }

   if (output->clienttags) {
      for (i = 0; i < ARGUS_MAXLISTEN; i++) {
         if ((output->clients[i].fd != -1) && (output->clients[i].sock != NULL)) {
            ArgusDeleteSocket(output, output->clients[i].sock);
            output->clienttags &= ~(1 << i);
            output->clientnum--;
            output->clients[i].fd = -1;
         }
      }
   }

#ifdef ARGUSDEBUG
   ArgusDebug (9, "ArgusOutputProcess() exiting\n");
#endif

#if defined(ARGUS_THREADS)
   pthread_exit(retn);
#endif /* ARGUS_THREADS */

   return (retn);
}


void
ArgusCheckClientStatus (struct ArgusOutputStruct *output)
{
   struct sockaddr from;
   int len = sizeof (from);
   int fd, i;

   if ((fd = accept (output->ArgusLfd, (struct sockaddr *)&from, (socklen_t *)&len)) > 0) {
      int flags = fcntl (fd, F_GETFL, 0L);
      if ((fcntl (fd, F_SETFL, flags | O_NONBLOCK)) >= 0) {
         if (ArgusTcpWrapper (fd, &from) >= 0) {
            for (i = 0; i < ARGUS_MAXLISTEN; i++) {
               if (!(output->clienttags & (1 << i))) {
                  bzero((char *)&output->clients[i], sizeof(output->clients[i]));
                  output->clienttags |= (1 << i);
                  output->clients[i].fd = fd;
                  output->clientnum++;
#ifdef ARGUSDEBUG
                  ArgusDebug (2, "ArgusCheckClientStatus() new client %d\n", i);
#endif
                  if ((output->clients[i].sock = ArgusNewSocket(fd)) == NULL)
                     ArgusLog (LOG_ERR, "ArgusInitOutput: ArgusNewSocket error %s", strerror(errno));

                  if (output->ArgusInitMar != NULL)
                     ArgusFree(output->ArgusInitMar);

                  if ((output->ArgusInitMar = ArgusGenerateInitialMar(output)) == NULL)
                     ArgusLog (LOG_ERR, "ArgusInitOutput: ArgusGenerateInitialMar error %s", strerror(errno));

                  len = ntohs(output->ArgusInitMar->hdr.len) * 4;
                  if (write (output->clients[i].fd, (char *) output->ArgusInitMar, len) != len) {
                     close (output->clients[i].fd);
                     ArgusLog (LOG_ERR, "ArgusInitOutput: write(): %s", strerror(errno));
                  }

                  break;
               }
            }

            if (i == ARGUS_MAXLISTEN) {
               char buf[MAXARGUSRECORD];
               struct ArgusRecord *argus = (struct ArgusRecord *) &buf;
               if ((argus = ArgusGenerateRecord (NULL, ARGUS_ERROR, (char *) &buf)) != NULL) {
                  len = ntohs(argus->hdr.len)  * 4;
                  argus->hdr.cause |= ARGUS_MAXLISTENEXCD;
                  if (write (fd, (char *) argus, len) != len)
                     ArgusLog (LOG_ERR, "ArgusInitOutput: write(): %s", strerror(errno));
                  close(fd);
               }
            }

         } else {
            ArgusLog (LOG_WARNING, "ArgusCheckClientStatus: ArgusTcpWrapper rejects");
            close (fd);
         }
         
      } else {
         ArgusLog (LOG_WARNING, "ArgusCheckClientStatus: fcntl: %s", strerror(errno));
         close (fd);
      }
   } else {
      ArgusLog (LOG_WARNING, "ArgusCheckClientStatus: accept: %s", strerror(errno));
      close (fd);
   }
     
#ifdef ARGUSDEBUG
   ArgusDebug (2, "ArgusCheckClientStatus() returning\n");
#endif
}


 
#define ARGUSMAXCLIENTCOMMANDS           6
#define RADIUM_START                    0
#define RADIUM_DONE                     1
#define RADIUM_FILTER                   2
#define RADIUM_MODEL                    3
#define RADIUM_PROJECT                  4
#define RADIUM_FILE                     5
 
char *ArgusClientCommands[ARGUSMAXCLIENTCOMMANDS] =
{
   "START: ",
   "DONE: ",
   "FILTER: ",
   "MODEL: ",
   "PROJECT: ",
   "FILE: ",
};


int
ArgusCheckClientMessage (struct ArgusOutputStruct *output, int ind)
{
   struct ArgusClientData *client = &output->clients[ind];
   int retn = 0, cnt = 0, i, found, fd = output->clients[ind].fd;
   char buf[MAXSTRLEN], *ptr = buf;
   unsigned int value = 0;
    
#ifdef ARGUS_SASL
   unsigned int *pvalue = &value;
   char *outputbuf = NULL;
   unsigned int outputlen = 0;

   if (client->sasl_conn)
      if ((retn = sasl_getprop(client->sasl_conn, SASL_MAXOUTBUF, (void **) &pvalue)) != SASL_OK)
         ArgusLog (LOG_ERR, "ArgusCheckClientMessage(0x%x, %d) sasl_getprop %s\n", client, fd, strerror(errno));
#endif /* ARGUS_SASL */

   bzero(buf, MAXSTRLEN);

   if (value == 0)
      value = MAXSTRLEN;

   if ((cnt = recv (fd, buf, value, 0)) <= 0) {
      if (cnt < 0) {
#ifdef ARGUSDEBUG
         ArgusDebug (3, "ArgusCheckClientMessage (0x%x, %d) recv() returned error %s\n", client, fd, strerror(errno));
#endif
         return (-1);

      } else {
#ifdef ARGUSDEBUG
         ArgusDebug (3, "ArgusCheckClientMessage (0x%x, %d) recv() returned %d bytes\n", client, fd, cnt);
#endif
         return(-3);
      }

   } else {
#ifdef ARGUSDEBUG
      ArgusDebug (3, "ArgusCheckClientMessage (0x%x, %d) recv() returned %d bytes\n", client, fd, cnt);
#endif
   }

#ifdef ARGUS_SASL
   if ((client->sasl_conn)) {
      if (sasl_decode (client->sasl_conn, buf, cnt, &outputbuf, &outputlen) != SASL_OK) {
         ArgusLog (LOG_WARNING, "ArgusCheckClientMessage(0x%x, %d) sasl_decode (0x%x, 0x%x, %d, 0x%x, %d) failed",
                    client, fd, client->sasl_conn, buf, cnt, &outputbuf, outputlen);
         return(-1);
      } else
#ifdef ARGUSDEBUG
         ArgusDebug (3, "ArgusCheckClientMessage (0x%x, %d) sasl_decode() returned %d bytes\n", client, fd, outputlen);
#endif
      if (outputlen > 0) {
         if (outputlen < MAXSTRLEN) {
            bzero (buf, MAXSTRLEN);
            bcopy (outputbuf, buf, outputlen);
            cnt = outputlen;
         } else
            ArgusLog (LOG_ERR, "ArgusCheckClientMessage(0x%x, %d) sasl_decode returned %d bytes\n", client, fd, outputlen);
        
         free(outputbuf);

      } else {
         return (0);
      }
   }

#endif /* ARGUS_SASL */

#ifdef ARGUSDEBUG
   ArgusDebug (3, "ArgusCheckClientMessage (0x%x, %d) read '%s' from remote\n", client, fd, ptr);
#endif

   for (i = 0, found = 0; i < ARGUSMAXCLIENTCOMMANDS; i++) {
      if (!(strncmp (ptr, ArgusClientCommands[i], strlen(ArgusClientCommands[i])))) {
         found++;
         switch (i) {
            case RADIUM_START: client->ArgusClientStart++; retn = 0; break;
            case RADIUM_DONE:  retn = -4; break; 
            case RADIUM_FILTER: {
#ifdef ARGUSDEBUG
               ArgusDebug (3, "ArgusCheckClientMessage: ArgusFilter %s\n", &ptr[7]);
#endif
               if (ArgusFilterCompile (&client->ArgusBPFcode, &ptr[7], 1) < 0) {
                  retn = -2;
#ifdef ARGUSDEBUG
                  ArgusDebug (3, "ArgusCheckClientMessage: ArgusFilter syntax error: %s\n", &ptr[7]);
#endif
               } else {
#ifdef ARGUSDEBUG
                  ArgusDebug (3, "ArgusCheckClientMessage: ArgusFilter %s\n", &ptr[7]);
#endif
                  client->ArgusFilterInitialized++;
                  if ((cnt = send (fd, "OK", 2, 0)) != 2) {
                     retn = -3;
#ifdef ARGUSDEBUG
                     ArgusDebug (3, "ArgusCheckClientMessage: send error %s\n", strerror(errno));
#endif
                  } else {
                     retn = 0;
#ifdef ARGUSDEBUG
                     ArgusDebug (3, "ArgusCheckClientMessage: ArgusFilter %s initialized.\n", &ptr[7]);
#endif
                  }
               }
               break;
            }

            case RADIUM_PROJECT: 
            case RADIUM_MODEL: 
               break;

            case RADIUM_FILE: {
               char *file = &ptr[6];
#ifdef ARGUSDEBUG
               ArgusDebug (3, "ArgusCheckClientMessage: ArgusFile %s requested.\n", file);
#endif
               RadiumSendFile (output, client, file, 0);
               retn = -5;
               break;
            }

            default:
               ArgusLog (LOG_WARNING, "ArgusCheckClientMessage: received %s",  ptr);
               break;
         }

         break;
      }
   }

   if (!found)
      ArgusLog (LOG_WARNING, "ArgusCheckClientMessage: received %s",  ptr);

#ifdef ARGUSDEBUG
   ArgusDebug (3, "ArgusCheckClientMessage(0x%x, %d): returning %d\n", output, ind, retn);
#endif

   return (retn);
}



/*
struct ArgusRecordHeader {
   unsigned int type:4;
   unsigned int cause:4;
   unsigned int vers:4;
   unsigned int opt:4;
   u_int16_t len;
};
 
struct ArgusRecord {
   struct ArgusRecordHeader hdr;
   union {
      struct ArgusMarStruct mar;
      struct ArgusFarStruct far;
   } ar_un;
};

struct ArgusMarStruct {
   struct timeval startime, now;
   unsigned char  major_version, minor_version;
   unsigned char interfaceType, interfaceStatus;
   unsigned short reportInterval, argusMrInterval;
   unsigned int argusid, localnet, netmask, nextMrSequenceNum;
   unsigned long long pktsRcvd, bytesRcvd;
   unsigned int pad[13], record_len;
};
*/


struct ArgusRecord *
ArgusGenerateInitialMar (struct ArgusOutputStruct *output)
{
   struct ArgusRecord *retn;
   struct timeval tbuf;

   if ((retn = (struct ArgusRecord *) ArgusCalloc (1, sizeof(struct ArgusRecord))) == NULL)
     ArgusLog (LOG_ERR, "ArgusGenerateInitialMar(0x%x) ArgusCalloc error %s\n", output, strerror(errno));
   
   retn->hdr.type  = ARGUS_MAR | ARGUS_VERSION;
   retn->hdr.cause = ARGUS_START;
   retn->hdr.len   = htons((unsigned short) sizeof(struct ArgusRecord)/4);

   retn->argus_mar.argusid = htonl(ARGUS_COOKIE);

   retn->argus_mar.startime.tv_sec  = output->ArgusStartTime.tv_sec;
   retn->argus_mar.startime.tv_usec = output->ArgusStartTime.tv_usec;

   gettimeofday (&tbuf, 0L);

   retn->argus_mar.now.tv_sec  = tbuf.tv_sec;
   retn->argus_mar.now.tv_usec = tbuf.tv_usec;

   retn->argus_mar.major_version = VERSION_MAJOR;
   retn->argus_mar.minor_version = VERSION_MINOR;
   retn->argus_mar.reportInterval = 0;
   retn->argus_mar.argusMrInterval = output->ArgusMarReportInterval.tv_sec;

   retn->argus_mar.localnet = output->ArgusLocalNet;
   retn->argus_mar.netmask = output->ArgusNetMask;

   retn->argus_mar.nextMrSequenceNum = output->ArgusOutputSequence;
   retn->argus_mar.thisid = getArgusID(ArgusParser);
   retn->argus_mar.record_len = -1;

#ifdef ARGUSDEBUG
   ArgusDebug (4, "ArgusGenerateInitialMar() returning\n");
#endif

   return (retn);
}

/*
struct ArgusRecordStruct *
ArgusGenerateStatusMarRecord (struct ArgusOutputStruct *output, unsigned char status)
{
   struct ArgusRecordStruct *retn;
   struct ArgusRecord *rec;
   struct timeval tbuf;

   if ((retn = (struct ArgusRecordStruct *) ArgusMallocListRecord(sizeof(*retn))) == NULL)
     ArgusLog (LOG_ERR, "ArgusGenerateStatusMarRecord(0x%x) ArgusCalloc error %s\n", output, strerror(errno));

   bzero ((char *) retn, sizeof(*retn));

   retn->hdr.type  = ARGUS_MAR | ARGUS_VERSION;
   retn->hdr.cause = status;
   retn->hdr.len   = ((unsigned short) sizeof(struct ArgusRecord)/4);

   rec = (struct ArgusRecord *) &retn->canon;

   rec->hdr = retn->hdr;
   rec->argus_mar.argusid = getArgusID(ArgusParser);

   rec->argus_mar.startime.tv_sec  = output->ArgusStartTime.tv_sec;
   rec->argus_mar.startime.tv_usec = output->ArgusStartTime.tv_usec;

   gettimeofday (&tbuf, 0L);

   rec->argus_mar.now.tv_sec  = tbuf.tv_sec;
   rec->argus_mar.now.tv_usec = tbuf.tv_usec;

   rec->argus_mar.major_version = VERSION_MAJOR;
   rec->argus_mar.minor_version = VERSION_MINOR;
   rec->argus_mar.reportInterval = 0;
   rec->argus_mar.argusMrInterval = output->ArgusMarReportInterval.tv_sec;

   rec->argus_mar.localnet = output->ArgusLocalNet;
   rec->argus_mar.netmask = output->ArgusNetMask;

   rec->argus_mar.nextMrSequenceNum = output->ArgusOutputSequence;
   rec->argus_mar.record_len = -1;

#ifdef ARGUSDEBUG
   ArgusDebug (4, "ArgusGenerateStatusMarRecord(0x%x) returning", output);
#endif

   return (retn);
}
*/


struct ArgusRecordStruct *
ArgusGenerateStatusMarRecord (struct ArgusOutputStruct *output, unsigned char status)
{
   extern int ArgusAllocTotal, ArgusFreeTotal, ArgusAllocBytes;
   struct ArgusRecordStruct *retn;
   struct ArgusRecord *rec;
   struct timeval now;


   if ((retn = (struct ArgusRecordStruct *) ArgusMallocListRecord (sizeof(*retn))) == NULL)
     ArgusLog (LOG_ERR, "ArgusGenerateStatusMarRecord(0x%x) ArgusCalloc error %s\n", output, strerror(errno));

   retn->hdr.type  = ARGUS_MAR | ARGUS_VERSION;
   retn->hdr.cause = status;
   retn->hdr.len   = ((unsigned short) sizeof(struct ArgusRecord)/4);

   rec = (struct ArgusRecord *) &retn->canon;

   rec->hdr = retn->hdr;
   rec->argus_mar.argusid = getArgusID(ArgusParser);

   gettimeofday (&now, 0L);

   rec->argus_mar.startime.tv_sec  = output->ArgusStartTime.tv_sec;
   rec->argus_mar.startime.tv_usec = output->ArgusStartTime.tv_usec;

   rec->argus_mar.now.tv_sec  = now.tv_sec;
   rec->argus_mar.now.tv_usec = now.tv_usec;

   rec->argus_mar.major_version = VERSION_MAJOR;
   rec->argus_mar.minor_version = VERSION_MINOR;
   rec->argus_mar.reportInterval = 0;
   rec->argus_mar.argusMrInterval = output->ArgusMarReportInterval.tv_sec;

   rec->argus_mar.localnet = 0;
   rec->argus_mar.netmask = 0;
 
   rec->argus_mar.nextMrSequenceNum = output->ArgusOutputSequence;
   rec->argus_mar.record_len = -1;

/*
   if ((ArgusSrc = output->ArgusSrc) != NULL) {
      int i;
      rec->argus_mar.interfaceType = ArgusSrc->ArgusInterface[0].ArgusInterfaceType;
      rec->argus_mar.interfaceStatus = getArgusInterfaceStatus(ArgusSrc);

      rec->argus_mar.pktsRcvd  = 0;
      rec->argus_mar.bytesRcvd = 0;
      rec->argus_mar.dropped   = 0;

      for (i = 0; i < ARGUS_MAXINTERFACE; i++) {
         rec->argus_mar.pktsRcvd  += ArgusSrc->ArgusInterface[i].ArgusStat.ps_recv - 
                                    ArgusSrc->ArgusInterface[i].ArgusLastPkts;
         rec->argus_mar.bytesRcvd += ArgusSrc->ArgusInterface[i].ArgusTotalBytes -
                                    ArgusSrc->ArgusInterface[i].ArgusLastBytes;
         rec->argus_mar.dropped   += ArgusSrc->ArgusInterface[i].ArgusStat.ps_drop - 
                                    ArgusSrc->ArgusInterface[i].ArgusLastDrop;

         ArgusSrc->ArgusInterface[i].ArgusLastPkts  = ArgusSrc->ArgusInterface[i].ArgusStat.ps_recv;
         ArgusSrc->ArgusInterface[i].ArgusLastDrop  = ArgusSrc->ArgusInterface[i].ArgusStat.ps_drop;
         ArgusSrc->ArgusInterface[i].ArgusLastBytes = ArgusSrc->ArgusInterface[i].ArgusTotalBytes;
      }
   }
*/

   rec->argus_mar.records = output->ArgusTotalRecords - output->ArgusLastRecords;
   output->ArgusLastRecords = output->ArgusTotalRecords;

/*
   rec->argus_mar.flows = output->ArgusModel->ArgusTotalNewFlows - output->ArgusModel->ArgusLastNewFlows;
   output->ArgusModel->ArgusLastNewFlows = output->ArgusModel->ArgusTotalNewFlows;

   if (output->ArgusModel && output->ArgusModel->ArgusStatusQueue)
      rec->argus_mar.queue   = output->ArgusModel->ArgusStatusQueue->count;
   else
      rec->argus_mar.queue   = 0;
*/

   if (output->ArgusOutputList)
      rec->argus_mar.output  = output->ArgusOutputList->count;
   else
      rec->argus_mar.output  = 0;

   rec->argus_mar.clients = output->clientnum;

   rec->argus_mar.bufs  = ArgusAllocTotal - ArgusFreeTotal;
   rec->argus_mar.bytes = ArgusAllocBytes;

#ifdef ARGUSDEBUG
   ArgusDebug (4, "ArgusGenerateStatusMarRecord(0x%x, %d) returning 0x%x", output, status, retn);
#endif

   return (retn);
}


#if defined HAVE_TCP_WRAPPER

#include <syslog.h>
#include <tcpd.h>

#ifndef MAXPATHNAMELEN
#define MAXPATHNAMELEN   BUFSIZ
#endif

#define PARANOID		1
#define KILL_IP_OPTIONS		1
#define HOSTS_ACCESS		1

int allow_severity = LOG_INFO;     /* run-time adjustable */
int deny_severity  = LOG_WARNING;   /* ditto */

void fix_options(struct request_info *);

#endif


int
ArgusTcpWrapper (int fd, struct sockaddr *from)
{
#if defined(HAVE_TCP_WRAPPER)
   int retn = 0;
   struct request_info request;

   /*
    * Find out the endpoint addresses of this conversation. Host name
    * lookups and double checks will be done on demand.
    */
 
   request_init(&request, RQ_DAEMON, ArgusParser->ArgusProgramName, RQ_FILE, STDIN_FILENO, 0);
   request.fd = fd;
   fromhost(&request);

   /*
    * Optionally look up and double check the remote host name. Sites
    * concerned with security may choose to refuse connections from hosts
    * that pretend to have someone elses host name.
    */
 
#ifdef PARANOID
   if (STR_EQ(eval_hostname(request.client), paranoid)) {
      ArgusLog (deny_severity, "refused connect from %s", eval_client(&request)); 
      if (request.sink)
         request.sink(request.fd);
      return -1;
   }
#endif

    /*
     * The BSD rlogin and rsh daemons that came out after 4.3 BSD disallow
     * socket options at the IP level. They do so for a good reason.
     * Unfortunately, we cannot use this with SunOS 4.1.x because the
     * getsockopt() system call can panic the system.
     */  

#if defined(KILL_IP_OPTIONS)
   fix_options(&request);
#endif /* KILL_IP_OPTIONS */

    /*
     * Find out and verify the remote host name. Sites concerned with
     * security may choose to refuse connections from hosts that pretend to
     * have someone elses host name.
     */  

#ifdef HOSTS_ACCESS
   if (!hosts_access(&request)) {
      ArgusLog  (deny_severity, "refused connect from %s", eval_client(&request));
      if (request.sink)
         request.sink(request.fd);
      return -1;
   } else
#endif

    /* Report remote client */
   ArgusLog  (allow_severity, "connect from %s", eval_client(&request));
   return (retn);

#else
   return (1);
#endif /* HAVE_TCP_WRAPPER */
}


struct ArgusSocketStruct *
ArgusNewSocket (int fd)
{
   struct ArgusSocketStruct *retn = NULL;

   if ((retn = ((struct ArgusSocketStruct *) ArgusCalloc (1, sizeof (struct ArgusSocketStruct)))) != NULL) {
      if ((retn->ArgusOutputList = ArgusNewList()) != NULL) {
         int flags = fcntl (fd, F_GETFL, 0L);
         fcntl (fd, F_SETFL, flags | O_NONBLOCK);
         retn->fd = fd;

      } else {
         ArgusFree(retn);   
         retn = NULL;
      }
   }

#ifdef ARGUSDEBUG
   ArgusDebug (2, "ArgusNewSocket (%d) returning 0x%x\n", fd, retn);
#endif

   return (retn);
}

void
ArgusDeleteSocket (struct ArgusOutputStruct *output, struct ArgusSocketStruct *asock)
{
   struct ArgusListStruct *list = asock->ArgusOutputList;
   struct ArgusListRecord *rec = NULL;   

   while (!(ArgusListEmpty (list)))
      if (ArgusWriteOutSocket(output, asock) < 0)
         break;

   while ((rec = (struct ArgusListRecord *) ArgusPopFrontList (list, ARGUS_LOCK)) != NULL)
      ArgusFree(rec);

   ArgusDeleteList(asock->ArgusOutputList);

   close(asock->fd);

   if (asock->filename)
      free(asock->filename);

   ArgusFree (asock);

#ifdef ARGUSDEBUG
   ArgusDebug (2, "ArgusDeleteSocket (0x%x) returning\n", asock);
#endif
}


#include <sys/stat.h>
#include <fcntl.h>

#define ARGUS_MAXERROR		20000
#define ARGUS_MAXWRITENUM	1024

int ArgusMaxListLength = 100000;
int ArgusCloseFile = 0;

int
ArgusWriteSocket (struct ArgusOutputStruct *output, struct ArgusSocketStruct *asock, struct ArgusRecordStruct *rec, int count)
{
   struct ArgusListStruct *list = asock->ArgusOutputList;
   struct stat statbuf;
   int retn = -1, ocnt;

   gettimeofday (&output->ArgusGlobalTime, 0L);

#ifdef ARGUSDEBUG
   ArgusDebug (6, "ArgusWriteSocket: called (0x%x, 0x%x, 0x%x, %d)\n", output, asock, rec, count);
#endif

   if (asock->fd != -1) {
      if (ArgusListEmpty (list) && (asock->rec == NULL)) {
         if (!(output->ArgusWriteStdOut) && (asock->filename)) {
            if (asock->lastwrite.tv_sec < output->ArgusGlobalTime.tv_sec) {
               if (((stat (asock->filename, &statbuf)) < 0) || (ArgusCloseFile)) {
                  close(asock->fd);
                  if ((asock->fd = open (asock->filename, O_WRONLY|O_APPEND|O_CREAT|O_NONBLOCK, 0x1a4)) < 0) {
                     ArgusLog (LOG_ERR, "ArgusWriteSocket: open(%s, O_WRONLY|O_APPEND|O_CREAT|O_NONBLOCK, 0x1a4) failed %s\n",
                                asock->filename, strerror(errno));
                  }
#ifdef ARGUSDEBUG
                  ArgusDebug (2, "ArgusWriteSocket: created outfile %s\n", asock->filename);
#endif

                  output->ArgusInitMar = ArgusGenerateInitialMar(output);
                  ocnt = sizeof(struct ArgusRecord);
                  if (((retn = write (asock->fd, output->ArgusInitMar, ocnt))) < ocnt)
                     ArgusLog (LOG_ERR, "ArgusWriteSocket: write %s failed %s\n", asock->filename, strerror(errno));
                  ArgusFree(output->ArgusInitMar);
                  output->ArgusInitMar = NULL;
               }

               asock->lastwrite = output->ArgusGlobalTime;
            }
         }

         if (ArgusGenerateRecord (rec, 0, (char *)&asock->buf)) {
            int cnt = ((struct ArgusRecord *)&asock->buf)->hdr.len * 4;
            
#if defined(_LITTLE_ENDIAN)
            ArgusHtoN((struct ArgusRecord *)&asock->buf);
#endif
            if (((retn = write (asock->fd, &asock->buf, cnt)) < cnt)) {
               if (retn < 0) {
                  if ((errno == EAGAIN) || (errno == EINTR) || 
                     ((errno == EPIPE) && !(asock->status & ARGUS_WAS_FUNCTIONAL)))
                     retn = 0;
                  else {
                     return (retn);
                  }
               }
            } else
               asock->status |= ARGUS_WAS_FUNCTIONAL;

            if (retn != cnt) {
               asock->writen = retn;
               asock->length = cnt;
               if (count > 1)
                  asock->rec = ArgusCopyRecordStruct(rec);
               else
                  asock->rec = rec;
            } else
               if (!(count > 1))
                  ArgusFreeListRecord(rec);
         }

      } else {
         if (list->count >= ArgusMaxListLength) {
#ifdef ARGUSDEBUG
            ArgusDebug (6, "ArgusWriteSocket: queue full\n");
#endif 
            if (ArgusWriteOutSocket(output, asock) < 0) {
               if (list->count >= ArgusMaxListLength) {
                  struct ArgusRecordStruct *trec;
                  int i;

#define ARGUS_DROPRECORDNUM  256
#ifdef ARGUSDEBUG
                  ArgusDebug (5, "ArgusWriteSocket: ArgusWriteOutSocket tossing records\n");
#endif 

                  for (i = 0; i < ARGUS_DROPRECORDNUM; i++)
                     if ((trec = (struct ArgusRecordStruct *) ArgusPopFrontList(list, ARGUS_LOCK)) != NULL)
                        ArgusFreeListRecord(trec);
               }
            }
         }

         if (count > 1)
            ArgusPushBackList (list, (struct ArgusListRecord *) ArgusCopyRecordStruct(rec), ARGUS_LOCK);
         else
            ArgusPushBackList (list, (struct ArgusListRecord *) rec, ARGUS_LOCK);

         retn = 0;
      }

   } else {
#ifdef ARGUSDEBUG
      ArgusDebug (4, "ArgusWriteSocket (0x%x, 0x%x, 0x%x) fd == %d\n", output, asock, rec);
#endif
   }

#ifdef ARGUSDEBUG
   ArgusDebug (6, "ArgusWriteSocket (0x%x, 0x%x, 0x%x) returning %d\n", output, asock, rec);
#endif

   return (retn);
}


#define ARGUS_LISTREPORTLEN	50000
#define ARGUS_LISTREPORTTIME	30

int
ArgusWriteOutSocket (struct ArgusOutputStruct *output, struct ArgusSocketStruct *asock)
{
   struct ArgusListStruct *list = NULL;
   struct ArgusRecordStruct *rec = NULL;
   int retn = 0, count = 1, len, cnt = 0;
   struct stat statbuf;
   unsigned char *ptr;

   if ((list = asock->ArgusOutputList) != NULL) {
      if ((count = ArgusGetListCount(list)) > 0) {
         if (count > ARGUS_MAXWRITENUM)
            count = ARGUS_MAXWRITENUM;
      }
   }

   while ((asock->fd != -1 ) && count--) {
      if ((rec = asock->rec) != NULL) {
         ptr = (unsigned char *)&asock->buf;
         if (!(asock->writen)) {
            if (asock->filename) {
               if (asock->lastwrite.tv_sec < output->ArgusGlobalTime.tv_sec) {
                  if (((stat (asock->filename, &statbuf)) < 0) || (ArgusCloseFile)) {
                     close(asock->fd);
                     if ((asock->fd = open (asock->filename, O_WRONLY|O_APPEND|O_CREAT|O_NONBLOCK, 0x1a4)) < 0) {
                        ArgusLog (LOG_ERR, "ArgusWriteOutSocket: open(%s, O_WRONLY|O_APPEND|O_CREAT|O_NONBLOCK, 0x1a4) failed %s\n",
                                   asock->filename, strerror(errno));
                     }
#ifdef ARGUSDEBUG
                     ArgusDebug (2, "ArgusWriteOutSocket: created outfile %s\n", asock->filename);
#endif
    
                     output->ArgusInitMar = ArgusGenerateInitialMar(output);
                     cnt = sizeof (struct ArgusRecord);
                     if (((retn = write (asock->fd, output->ArgusInitMar, cnt))) < cnt)
                        ArgusLog (LOG_ERR, "ArgusWriteOutSocket: write %s failed %s\n", asock->filename, strerror(errno));
                     ArgusFree(output->ArgusInitMar); 
                     output->ArgusInitMar = NULL;
                  }

                  asock->lastwrite = output->ArgusGlobalTime;
               }
            }
         }
         
         if ((asock->writen < asock->length) && ( asock->writen >= 0)) {
            len = asock->length - asock->writen;

            if ((retn = write(asock->fd, (unsigned char *)&ptr[asock->writen], len)) >= 0) {
               asock->errornum = 0;
               asock->writen += retn;

            } else {
               if ((retn < 0) && !((errno == EAGAIN) || (errno == EINTR))) {
                  ArgusLog (LOG_WARNING, "ArgusWriteOutSocket: write() failed\n", strerror(errno));
                  asock->errornum++;
               } else {
                  if (asock->errornum++ < ARGUS_MAXERROR) {
                     retn = 0;
                  }
               }
               break;
            }
         }
         
         if (asock->writen >= asock->length) {
            gettimeofday(&list->outputTime, 0L);
            ArgusFreeListRecord(rec);

            if ((rec = (struct ArgusRecordStruct *) ArgusPopFrontList(list, ARGUS_LOCK)) != NULL) {
               if (ArgusGenerateRecord (rec, 0, (char *)&asock->buf)) {
                  int cnt = ((struct ArgusRecord *)&asock->buf)->hdr.len * 4;
#if defined(_LITTLE_ENDIAN)
                  ArgusHtoN((struct ArgusRecord *)&asock->buf);
#endif
                  asock->writen = 0;
                  asock->length = cnt;
                  asock->rec = rec;
               } else {
                  ArgusFreeListRecord(rec);
                  count--;
               }
            }

         } else {
            asock->rec = rec;
            break;
         }
      }
   }

#ifdef ARGUSDEBUG
   ArgusDebug (9, "ArgusWriteOutSocket(0x%x): queue empty\n", asock);
#endif
   
   if (asock->errornum >= ARGUS_MAXERROR) {
      ArgusLog (LOG_WARNING, "ArgusWriteOutSocket(0x%x) maximum errors exceeded %d\n",
                         asock, asock->errornum);
      retn = -1;
   }

   if ((count = ArgusGetListCount(list)) > ArgusMaxListLength) {
      ArgusLog (LOG_WARNING, "ArgusWriteOutSocket(0x%x) max queue exceeded %d\n",
                         asock, count);
      retn = -1;
   }

#ifdef ARGUSDEBUG
   if (list)
      ArgusDebug (8, "ArgusWriteOutSocket (0x%x) %d records waiting. returning %d\n", asock, list->count, retn);
   else
      ArgusDebug (8, "ArgusWriteOutSocket (0x%x) no list.  returning %d\n", asock, list->count, retn);
#endif

   return retn;
}


void
RadiumSendFile (struct ArgusOutputStruct *output, struct ArgusClientData *client, char *file, int status)
{
   int retn = 0, pid = 0, cnt, fd = client->fd, flags, error = 0;
   unsigned int filesize;
   char sbuf[0x10000];
   struct stat statbuf;
   FILE *ffd = NULL;


   flags = fcntl (fd, F_GETFL, 0L);
   fcntl (fd, F_SETFL, flags & ~O_NONBLOCK);

   switch ((retn = stat (file, &statbuf))) {
      case 0:
         if (statbuf.st_mode & S_IFREG)
            break;

      case -1:
         error++;
         break;
   }

   if (error) {
      if ((cnt = send (fd, "KO", 2, 0)) != 2) {
#ifdef ARGUSDEBUG
         ArgusDebug (3, "RadiumSendFile: send error %s\n", strerror(errno));
#endif
      }
#ifdef ARGUSDEBUG
      ArgusDebug (3, "RadiumSendFile: file %s error %s", file, strerror(errno));
#endif
      client->pid = 0;
      return;
   }

   filesize = statbuf.st_size;
#ifdef ARGUSDEBUG
   ArgusDebug (3, "RadiumSendFile: %s size is %d", file, filesize);
#endif

   if ((pid = fork ()) < 0)
      ArgusLog (LOG_ERR, "Can't fork file processor %s", strerror(errno));

   if (!(pid)) {
#if defined(_LITTLE_ENDIAN)
         filesize = htonl(filesize);
#endif

         if ((cnt = send (fd, &filesize, sizeof(filesize), 0)) != sizeof(filesize)) {
#ifdef ARGUSDEBUG
            ArgusDebug (3, "RadiumSendFile: send error %s\n", strerror(errno));
#endif
         } else {
#ifdef ARGUSDEBUG
            ArgusDebug (3, "RadiumSendFile: sent %d bytes");
#endif
         }

         if ((cnt = recv (fd, sbuf, 0x10000, 0)) <= 0) {
            if (cnt < 0) {
               ArgusLog (LOG_ERR, "RadiumSendFile (0x%x, %d) recv() returned error %s\n", client, fd, strerror(errno));

            } else {
               ArgusLog (LOG_ERR, "RadiumSendFile (0x%x, %d) recv() returned %d bytes\n", client, fd, cnt);
            }
         }

         if (strstr (sbuf, "START")) {
#ifdef ARGUSDEBUG
            ArgusDebug (3, "RadiumSendFile: received START from requestor");
#endif
            if ((ffd = fopen (file, "r")) != NULL) {
               char sbuf[0x10000];
               int cnt = 0, bytes = 0;
               while ((cnt = fread (sbuf, 1, 0x10000, ffd)) > 0) {
                  if (write (fd, sbuf, cnt) < cnt) {
                     ArgusLog (LOG_ERR, "remote file transfer write error", strerror(errno));
                  }
                  bytes += cnt;
               }

               if (ferror(ffd))
                  ArgusLog (LOG_ERR, "local file transfer read error", strerror(errno));

#ifdef ARGUSDEBUG
               ArgusDebug (3, "RadiumSendFile: file %s sent %d bytes\n", file, bytes);
#endif
            }
         }

#ifdef ARGUSDEBUG
      ArgusDebug (3, "RadiumSendFile: file %s initialized.\n", file);
#endif

      exit (0);

   } else {
      client->pid = pid;
#ifdef ARGUSDEBUG
      ArgusDebug (3, "RadiumSendFile (0x%x, %s): forked file processor\n", client, file);
#endif
   }
}

